#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "main.h"
#include "Graphics.h"
#include "System.h"
#include "CSDL.h"
#include "math/Complex.h"
#include "Polynomial.h"

#define SCREEN_WIDTH    1280
#define SCREEN_HEIGHT   720
#define MAX_ITERATIONS  40
#define TOLERANCE       1e-6

Polynomial p;
Polynomial dp;

// the function
Complex f(Complex z)
{
    return p.evaluate(z);
}

// derivative of the function
Complex df(Complex z)
{
    return dp.evaluate(z);
}

static double zoom = 2.0 / (double)SCREEN_HEIGHT;
static double centerX = (double)SCREEN_WIDTH  / 2.0;
static double centerY = (double)SCREEN_HEIGHT / 2.0;

void drawNewton(int offsetX, int offsetY, std::vector<Color>& palette)
{
    for (int yy = 0; yy < SCREEN_HEIGHT; yy += 8)
    {
        for (int xx = 0; xx < SCREEN_WIDTH; xx += 8)
        {
            int x = xx + offsetX;
            int y = yy + offsetY;

            // convert the pixel pos to a complex number
            Complex z;
            z.re = (x - centerX) * zoom;
            z.im = (y - centerY) * zoom;

            // by default the pixel is black if the calculations fail
            int palIdx = p.roots.size();

            // iterate
            int iter;

            for (iter = 0; iter < MAX_ITERATIONS; iter++)
            {
                z = z - f(z) / df(z);

                // are we near a root ?
                for (int i = 0; i < p.roots.size(); i++)
                {
                    Complex delta = z - p.roots[i];
                    double distance = delta.squareMag();

                    if (distance < TOLERANCE)
                    {
                        palIdx = i;
                        break;
                    }
                }

                if (palIdx != p.roots.size())
                    break;
            }

            // draw the pixel
            Color color = palette[palIdx];

            if (palIdx != p.roots.size())
            {
                float coef = 1 - log(iter + 1) / 5.0;

                if (coef < 0)
                    coef = 0;

                color.r *= coef;
                color.g *= coef;
                color.b *= coef;
            }
            gfx.setPixel(x, y, color);
        }
    }
}

int main(int argc, char* argv[])
{
    // init the window
    gfx.init("Newton tool", SCREEN_WIDTH, SCREEN_HEIGHT);
    gfx.init2D();
    gfx.clearScreen(Color(0, 0, 0, SDL_ALPHA_OPAQUE));

    // palette
    std::vector<Color> palette;
    palette.push_back(Color(0, 0, 0));

    static Uint8 pattern[64] =
    {
         0, 36,  4, 32, 18, 54, 22, 50,
         2, 38,  6, 34, 16, 52, 20, 48,
         9, 45, 13, 41, 27, 63, 31, 59,
        11, 47, 15, 43, 25, 61, 29, 57,
         1, 37,  5, 33, 19, 55, 23, 51,
         3, 39,  7, 35, 17, 53, 21, 49,
         8, 44, 12, 40, 26, 62, 30, 58,
        10, 46, 14, 42, 24, 60, 28, 56
    };

    srand(time(NULL));
    int frameCount = 0;

    while (sys.isQuitRequested() == false)
    {
        // draw the fractal
        if (frameCount < 64)
        {
            drawNewton(pattern[frameCount] % 8,
                       pattern[frameCount] / 8,
                       palette);
            frameCount++;
        }

        // display and process the events
        gfx.render();
        sys.wait(20);
        sys.processEvents();

        // left mouse button: add a new root
        if (sys.buttonsDown & BUT_LEFT)
        {
            // restart the drawing
            frameCount = 0;

            // choose a random color
            palette[palette.size()-1] = Color(rand() % 256, rand() % 256, rand() % 256);
            palette.push_back(Color(0, 0, 0));

            // add the root
            Complex r;
            r.re = (sys.mouseX - centerX) * zoom;
            r.im = (sys.mouseY - centerY) * zoom;
            p.addRoot(r);

            // recompute the coefficients
            p.findCoefficients();
            p.derivative(dp);
        }

        // right mouse button: remove the last root
        if (sys.buttonsDown & BUT_RIGHT)
        {
            // if there is at least one root
            if (p.roots.size() > 0)
            {
                // restart the drawing
                frameCount = 0;

                // remove the root
                palette.pop_back();
                p.remRoot();

                // recompute the coefficients
                p.findCoefficients();
                p.derivative(dp);
            }
        }
    }

    gfx.quit();

    return EXIT_SUCCESS;
}
